"""
Some functions internally use by SortingComparison.
"""

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy.optimize import linear_sum_assignment


def count_matching_events(times1, times2, delta=10):
    """
    Counts matching events.

    Parameters
    ----------
    times1: list
        List of spike train 1 frames
    times2: list
        List of spike train 2 frames
    delta: int
        Number of frames for considering matching events

    Returns
    -------
    matching_count: int
        Number of matching events
    """
    times_concat = np.concatenate((times1, times2))
    membership = np.concatenate((np.ones(times1.shape) * 1, np.ones(times2.shape) * 2))
    indices = times_concat.argsort()
    times_concat_sorted = times_concat[indices]
    membership_sorted = membership[indices]
    diffs = times_concat_sorted[1:] - times_concat_sorted[:-1]
    inds = np.where((diffs <= delta) & (membership_sorted[:-1] != membership_sorted[1:]))[0]
    if len(inds) == 0:
        return 0
    inds2 = np.where(inds[:-1] + 1 != inds[1:])[0]
    return len(inds2) + 1


def compute_agreement_score(num_matches, num1, num2):
    """
    Computes agreement score.

    Parameters
    ----------
    num_matches: int
        Number of matches
    num1: int
        Number of events in spike train 1
    num2: int
        Number of events in spike train 2

    Returns
    -------
    score: float
        Agreement score
    """
    denom = num1 + num2 - num_matches
    if denom == 0:
        return 0
    return num_matches / denom


def do_count_event(sorting):
    """
    Count event for each units in a sorting.
    Parameters
    ----------
    sorting: SortingExtractor
        A sorting extractor
    
    Returns
    -------
    event_count: pd.Series
        Nb of spike by units.
    """
    unit_ids = sorting.get_unit_ids()
    ev_counts = np.array([len(sorting.get_unit_spike_train(u)) for u in unit_ids], dtype='int64')
    event_counts = pd.Series(ev_counts, index=unit_ids)
    return event_counts


def count_match_spikes(times1, all_times2, delta_frames):  # , event_counts1, event_counts2  unit2_ids,
    """
    Computes matching spikes between one spike train and a list of others.

    Parameters
    ----------
    times1: array
        Spike train 1 frames
    all_times2: list of array
        List of spike trains from sorting 2

    Returns
    -------
    matching_events_count: list
        List of counts of matching events
    """
    matching_event_counts = np.zeros(len(all_times2), dtype='int64')
    for i2, times2 in enumerate(all_times2):
        num_matches = count_matching_events(times1, times2, delta=delta_frames)
        matching_event_counts[i2] = num_matches
    return matching_event_counts


def make_match_count_matrix(sorting1, sorting2, delta_frames, n_jobs=1):
    """
    Make the match_event_count matrix.
    Basically it count the match event for all given pair of spike train from
    sorting1 and sorting2.
    
    Parameters
    ----------
    sorting1: SortingExtractor
        The first sorting extractor

    sorting2: SortingExtractor
        The second sorting extractor

    delta_frames: int
        Number of frames to consider spikes coincident

    n_jobs: int
        Number of jobs to run in parallel

    Returns
    -------
    
    match_event_count: array (int64)
        Matrix of match count spike
    
    """
    unit1_ids = np.array(sorting1.get_unit_ids())
    unit2_ids = np.array(sorting2.get_unit_ids())

    # preload all spiketrains 2 into a list
    s2_spiketrains = [sorting2.get_unit_spike_train(u2) for u2 in unit2_ids]

    match_event_count_lists = Parallel(n_jobs=n_jobs)(delayed(count_match_spikes)(sorting1.get_unit_spike_train(u1),
                                                                                  s2_spiketrains, delta_frames) for
                                                      i1, u1 in enumerate(unit1_ids))

    match_event_count = pd.DataFrame(np.array(match_event_count_lists),
                                     index=unit1_ids, columns=unit2_ids)

    return match_event_count


def make_agreement_scores(sorting1, sorting2, delta_frames, n_jobs=1):
    """
    Make the agreement matrix.
    No threshold (min_score) is applied at this step.
    
    Note : this computation is symetric.
    Inverting sorting1 and sorting2 give the transposed matrix.
    

    Parameters
    ----------
    sorting1: SortingExtractor
        The first sorting extractor

    sorting2: SortingExtractor
        The second sorting extractor

    delta_frames: int
        Number of frames to consider spikes coincident

    n_jobs: int
        Number of jobs to run in parallel

    Returns
    -------
    
    agreement_scores: array (float)
        The agreement score matrix.
    
    """
    unit1_ids = np.array(sorting1.get_unit_ids())
    unit2_ids = np.array(sorting2.get_unit_ids())

    ev_counts1 = np.array([len(sorting1.get_unit_spike_train(u1)) for u1 in unit1_ids], dtype='int64')
    ev_counts2 = np.array([len(sorting2.get_unit_spike_train(u2)) for u2 in unit2_ids], dtype='int64')
    event_counts1 = pd.Series(ev_counts1, index=unit1_ids)
    event_counts2 = pd.Series(ev_counts2, index=unit2_ids)

    match_event_count = make_match_count_matrix(sorting1, sorting2, delta_frames, n_jobs=n_jobs)

    agreement_scores = make_agreement_scores_from_count(match_event_count, event_counts1, event_counts2)

    return agreement_scores


def make_agreement_scores_from_count(match_event_count, event_counts1, event_counts2):
    """
    See make_agreement_scores.
    Other signature here to avoid to recompute match_event_count matrix.
    
    Parameters
    ----------
    
    match_event_count
    
    
    """

    # numpy broadcast style
    denom = event_counts1.values[:, None] + event_counts2.values[None, :] - match_event_count.values
    # little trick here when denom is 0 to avoid 0 division : lets put -1
    # it will 0 anyway
    denom[denom == 0] = -1

    agreement_scores = match_event_count.values / denom
    agreement_scores = pd.DataFrame(agreement_scores,
                                    index=match_event_count.index, columns=match_event_count.columns)
    return agreement_scores


def make_possible_match(agreement_scores, min_score):
    """
    Given an agreement matrix and a min_score threshold.
    Return as a dict all possible match for each spiketrain in each side.
    
    Note : this is symmetric.
    

    Parameters
    ----------
    agreement_scores: pd.DataFrame
    
    min_score: float
    
    
    Returns
    -----------
    best_match_12: pd.Series
    
    best_match_21: pd.Series    
    
    """
    unit1_ids = np.array(agreement_scores.index)
    unit2_ids = np.array(agreement_scores.columns)

    # threshold the matrix
    scores = agreement_scores.values.copy()
    scores[scores < min_score] = 0

    possible_match_12 = {}
    for i1, u1 in enumerate(unit1_ids):
        inds_match, = np.nonzero(scores[i1, :])
        possible_match_12[u1] = unit2_ids[inds_match]

    possible_match_21 = {}
    for i2, u2 in enumerate(unit2_ids):
        inds_match, = np.nonzero(scores[:, i2])
        possible_match_21[u2] = unit1_ids[inds_match]

    return possible_match_12, possible_match_21


def make_best_match(agreement_scores, min_score):
    """
    Given an agreement matrix and a min_score threshold.
    return a dict a best match for each units independently of others.
    
    Note : this is symmetric.

    Parameters
    ----------
    agreement_scores: pd.DataFrame
    
    min_score: float
    
    
    Returns
    -----------
    best_match_12: pd.Series
    
    best_match_21: pd.Series
    
    
    """
    unit1_ids = np.array(agreement_scores.index)
    unit2_ids = np.array(agreement_scores.columns)

    scores = agreement_scores.values.copy()

    best_match_12 = pd.Series(index=unit1_ids, dtype='int64')
    best_match_12[:] = -1
    for i1, u1 in enumerate(unit1_ids):
        if scores.shape[1]>0: 
            ind_max = np.argmax(scores[i1, :])
            if scores[i1, ind_max] >= min_score:
                best_match_12[u1] = unit2_ids[ind_max]

    best_match_21 = pd.Series(index=unit2_ids, dtype='int64')
    best_match_21[:] = -1
    for i2, u2 in enumerate(unit2_ids):
        if scores.shape[0]>0: 
            ind_max = np.argmax(scores[:, i2])
            if scores[ind_max, i2] >= min_score:
                best_match_21[u2] = unit1_ids[ind_max]

    return best_match_12, best_match_21


def make_hungarian_match(agreement_scores, min_score):
    """
    Given an agreement matrix and a min_score threshold.
    return the "optimal" match with the "hungarian" algo.
    This use internally the scipy.optimze.linear_sum_assignment implementation.
    
    Parameters
    ----------
    agreement_scores: pd.DataFrame
    
    min_score: float
    
    
    Returns
    -----------
    hungarian_match_12: pd.Series
    
    hungarian_match_21: pd.Series
    
    """
    unit1_ids = np.array(agreement_scores.index)
    unit2_ids = np.array(agreement_scores.columns)

    # threshold the matrix
    scores = agreement_scores.values.copy()
    scores[scores < min_score] = 0

    [inds1, inds2] = linear_sum_assignment(-scores)

    hungarian_match_12 = pd.Series(index=unit1_ids, dtype='int64')
    hungarian_match_12[:] = -1
    hungarian_match_21 = pd.Series(index=unit2_ids, dtype='int64')
    hungarian_match_21[:] = -1

    for i1, i2 in zip(inds1, inds2):
        u1 = unit1_ids[i1]
        u2 = unit2_ids[i2]
        if agreement_scores.at[u1, u2] >= min_score:
            hungarian_match_12[u1] = u2
            hungarian_match_21[u2] = u1

    return hungarian_match_12, hungarian_match_21


def do_score_labels(sorting1, sorting2, delta_frames, unit_map12, label_misclassification=False):
    """
    Makes the labelling at spike level for each spike train:
      * TP: true positive
      * CL: classification error
      * FN: False negative
      * FP: False positive
      * TOT:
      * TOT_ST1:
      * TOT_ST2:

    Parameters
    ----------
    sorting1: SortingExtractor instance
        The ground truth sorting
    sorting2: SortingExtractor instance
        The tested sorting
    delta_frames: int
        Number of frames to consider spikes coincident
    unit_map12: pd.Series
        Dict of matching from sorting1 to sorting2
    label_misclassification: bool
        If True, misclassification errors are labelled

    Returns
    -------
    labels_st1: dict of np.array of str
        Contain score labels for units of sorting 1
    labels_st2: dict of np.array of str
        Contain score labels for units of sorting 2
    """
    unit1_ids = sorting1.get_unit_ids()
    unit2_ids = sorting2.get_unit_ids()
    labels_st1 = dict()
    labels_st2 = dict()
    N1 = len(unit1_ids)
    N2 = len(unit2_ids)

    # copy spike trains for faster access from extractors with memmapped data
    sts1 = {u1: sorting1.get_unit_spike_train(u1) for u1 in unit1_ids}
    sts2 = {u2: sorting2.get_unit_spike_train(u2) for u2 in unit2_ids}

    for u1 in unit1_ids:
        lab_st1 = np.array(['UNPAIRED'] * len(sts1[u1]), dtype='<U8')
        labels_st1[u1] = lab_st1
    for u2 in unit2_ids:
        lab_st2 = np.array(['UNPAIRED'] * len(sts2[u2]), dtype='<U8')
        labels_st2[u2] = lab_st2

    for u1 in unit1_ids:
        u2 = unit_map12[u1]
        if u2 != -1:
            lab_st1 = labels_st1[u1]
            lab_st2 = labels_st2[u2]
            mapped_st = sorting2.get_unit_spike_train(u2)
            times_concat = np.concatenate((sts1[u1], mapped_st))
            membership = np.concatenate((np.ones(sts1[u1].shape) * 1, np.ones(mapped_st.shape) * 2))
            indices = times_concat.argsort()
            times_concat_sorted = times_concat[indices]
            membership_sorted = membership[indices]
            diffs = times_concat_sorted[1:] - times_concat_sorted[:-1]
            inds = np.where((diffs <= delta_frames) & (membership_sorted[:-1] != membership_sorted[1:]))[0]
            if len(inds) > 0:
                inds2 = inds[np.where(inds[:-1] + 1 != inds[1:])[0]] + 1
                inds2 = np.concatenate((inds2, [inds[-1]]))
                times_matched = times_concat_sorted[inds2]
                # find and label closest spikes
                ind_st1 = np.array([np.abs(sts1[u1] - tm).argmin() for tm in times_matched])
                ind_st2 = np.array([np.abs(mapped_st - tm).argmin() for tm in times_matched])
                assert(len(np.unique(ind_st1)) == len(ind_st1))
                assert(len(np.unique(ind_st2)) == len(ind_st2))
                lab_st1[ind_st1] = 'TP'
                lab_st2[ind_st2] = 'TP'
        else:
            lab_st1 = np.array(['FN'] * len(sts1[u1]))
            labels_st1[u1] = lab_st1

    if label_misclassification:
        for u1 in unit1_ids:
            lab_st1 = labels_st1[u1]
            st1 = sts1[u1]
            for l_gt, lab in enumerate(lab_st1):
                if lab == 'UNPAIRED':
                    for u2 in unit2_ids:
                        if u2 in unit_map12.values and unit_map12[u1] != -1:
                            lab_st2 = labels_st2[u2]
                            n_sp = st1[l_gt]
                            mapped_st = sts2[u2]
                            matches = (np.abs(mapped_st.astype(int) - n_sp) <= delta_frames)
                            if np.sum(matches) > 0:
                                if 'CL' not in lab_st1[l_gt] and 'CL' not in lab_st2[np.where(matches)[0][0]]:
                                    lab_st1[l_gt] = 'CL_' + str(u1) + '_' + str(u2)
                                    lab_st2[np.where(matches)[0][0]] = 'CL_' + str(u2) + '_' + str(u1)

    for u1 in unit1_ids:
        lab_st1 = labels_st1[u1]
        lab_st1[lab_st1 == 'UNPAIRED'] = 'FN'

    for u2 in unit2_ids:
        lab_st2 = labels_st2[u2]
        lab_st2[lab_st2 == 'UNPAIRED'] = 'FP'

    return labels_st1, labels_st2


def compare_spike_trains(spiketrain1, spiketrain2, delta_frames=10):
    """
    Compares 2 spike trains.

    Note:
      * The first spiketrain is supposed to be the ground truth.
      * this implementation do not count a TP when more than one spike
        is present around the same time in spiketrain2.

    Parameters
    ----------
    spiketrain1,spiketrain2: numpy.array
        Times of spikes for the 2 spike trains.

    Returns
    -------
    lab_st1, lab_st2: numpy.array
        Label of score for each spike

    """
    lab_st1 = np.array(['UNPAIRED'] * len(spiketrain1))
    lab_st2 = np.array(['UNPAIRED'] * len(spiketrain2))

    # from gtst: TP, TPO, TPSO, FN, FNO, FNSO
    for sp_i, n_sp in enumerate(spiketrain1):
        matches = (np.abs(spiketrain2.astype(int) - n_sp) <= delta_frames // 2)
        if np.sum(matches) > 0:
            if lab_st1[sp_i] != 'TP' and lab_st2[np.where(matches)[0][0]] != 'TP':
                lab_st1[sp_i] = 'TP'
                lab_st2[np.where(matches)[0][0]] = 'TP'

    for l_gt, lab in enumerate(lab_st1):
        if lab == 'UNPAIRED':
            lab_st1[l_gt] = 'FN'

    for l_gt, lab in enumerate(lab_st2):
        if lab == 'UNPAIRED':
            lab_st2[l_gt] = 'FP'

    return lab_st1, lab_st2


def do_confusion_matrix(event_counts1, event_counts2, match_12, match_event_count):
    """
    Computes the confusion matrix between one ground truth sorting
    and another sorting.

    Parameters
    ----------
    event_counts1: pd.Series
        Number of event per units 1
    event_counts2: pd.Series
        Number of event per units 2
    match_12: pd.Series
        Series of matching from sorting1 to sorting2.
        Can be the hungarian or best match.
    match_event_count: pd.DataFrame
        The match count matrix given by make_match_count_matrix
        
    Returns
    ------
    confusion_matrix: pd.DataFrame
        The confusion matrix
        index are units1 reordered
        columns are units2 redordered
    """
    unit1_ids = np.array(match_event_count.index)
    unit2_ids = np.array(match_event_count.columns)
    N1 = len(unit1_ids)
    N2 = len(unit2_ids)

    matched_units1 = match_12[match_12 != -1].index
    matched_units2 = match_12[match_12 != -1].values

    unmatched_units1 = match_12[match_12 == -1].index
    unmatched_units2 = unit2_ids[~np.in1d(unit2_ids, matched_units2)]

    ordered_units1 = np.hstack([matched_units1, unmatched_units1])
    ordered_units2 = np.hstack([matched_units2, unmatched_units2])

    conf_matrix = pd.DataFrame(np.zeros((N1 + 1, N2 + 1), dtype=int),
                               index=list(ordered_units1) + ['FP'],
                               columns=list(ordered_units2) + ['FN'])

    for u1 in matched_units1:
        u2 = match_12[u1]
        num_match = match_event_count.at[u1, u2]
        conf_matrix.at[u1, u2] = num_match
        conf_matrix.at[u1, 'FN'] = event_counts1.at[u1] - num_match
        conf_matrix.at['FP', u2] = event_counts2.at[u2] - num_match

    for u1 in unmatched_units1:
        conf_matrix.at[u1, 'FN'] = event_counts1.at[u1]

    for u2 in unmatched_units2:
        conf_matrix.at['FP', u2] = event_counts2.at[u2]

    return conf_matrix


def do_count_score(event_counts1, event_counts2, match_12, match_event_count):
    """
    For each ground truth units count how many:
    'tp', 'fn', 'cl', 'fp', 'num_gt', 'num_tested', 'tested_id'
    
    Parameters
    ----------
    event_counts1: pd.Series
        Number of event per units 1
    event_counts2: pd.Series
        Number of event per units 2
    match_12: pd.Series
        Series of matching from sorting1 to sorting2.
        Can be the hungarian or best match.
    match_event_count: pd.DataFrame
        The match count matrix given by make_match_count_matrix
    Returns
    ----------
    count_score: pd.DataFrame
        A table with one line per GT units and columns
        are tp/fn/fp/...
    """

    unit1_ids = event_counts1.index

    columns = ['tp', 'fn', 'fp', 'num_gt', 'num_tested', 'tested_id']

    count_score = pd.DataFrame(index=unit1_ids, columns=columns)
    count_score.index.name = 'gt_unit_id'
    for i1, u1 in enumerate(unit1_ids):
        u2 = match_12[u1]
        count_score.at[u1, 'tested_id'] = u2
        if u2 == -1:
            count_score.at[u1, 'num_tested'] = 0
            count_score.at[u1, 'tp'] = 0
            count_score.at[u1, 'fp'] = 0
            count_score.at[u1, 'fn'] = event_counts1.at[u1]
            count_score.at[u1, 'num_gt'] = event_counts1.at[u1]
        else:
            num_match = match_event_count.at[u1, u2]
            count_score.at[u1, 'tp'] = num_match
            count_score.at[u1, 'fn'] = event_counts1.at[u1] - num_match
            count_score.at[u1, 'fp'] = event_counts2.at[u2] - num_match

            count_score.at[u1, 'num_gt'] = event_counts1.at[u1]
            count_score.at[u1, 'num_tested'] = event_counts2.at[u2]

    return count_score


_perf_keys = ['accuracy', 'recall', 'precision', 'false_discovery_rate', 'miss_rate']


def compute_performence(count_score):
    """
    This compte perf formula.
    this trick here is that it works both on pd.Series and pd.Dataframe
    line by line.
    This it is internally used by perf by psiketrain and poll_with_sum.
    
    https://en.wikipedia.org/wiki/Sensitivity_and_specificity
    
    Note :
      * we don't have TN because it do not make sens here.
      * 'accuracy' = 'tp_rate' because TN=0
      * 'recall' = 'sensitivity'
    """

    perf = pd.DataFrame(index=count_score.index, columns=_perf_keys)
    perf.index.name = 'gt_unit_id'
    perf[:] = 0

    # make it robust when num_gt is 0
    keep = (count_score['num_gt'] > 0) & (count_score['tp'] > 0)

    c = count_score.loc[keep]
    tp, fn, fp, num_gt = c['tp'], c['fn'], c['fp'], c['num_gt']

    perf.loc[keep, 'accuracy'] = tp / (tp + fn + fp)
    perf.loc[keep, 'recall'] = tp / (tp + fn)
    perf.loc[keep, 'precision'] = tp / (tp + fp)
    perf.loc[keep, 'false_discovery_rate'] = fp / (tp + fp)
    perf.loc[keep, 'miss_rate'] = fn / num_gt

    return perf


def make_matching_events(times1, times2, delta):
    """
    Similar to count_matching_events but get index instead of counting.
    Used for collision detection

    Parameters
    ----------
    times1: list
        List of spike train 1 frames
    times2: list
        List of spike train 2 frames
    delta: int
        Number of frames for considering matching events

    Returns
    -------
    matching_event: numpy array dtype = ['index1', 'index2', 'delta']
        1d of collision
        
        
    """
    times_concat = np.concatenate((times1, times2))
    membership = np.concatenate((np.ones(times1.shape) * 1, np.ones(times2.shape) * 2))
    spike_idx = np.concatenate((np.arange(times1.size, dtype='int64'), np.arange(times2.size, dtype='int64')))
    indices = times_concat.argsort()
    
    times_concat_sorted = times_concat[indices]
    membership_sorted = membership[indices]
    spike_index_sorted = spike_idx[indices]
    
    inds, = np.nonzero((np.diff(times_concat_sorted) <= delta) & (np.diff(membership_sorted) != 0))
    
    dtype = [('index1', 'int64'), ('index2', 'int64'), ('delta_frame', 'int64')]
    
    if len(inds) == 0:
        return np.array([], dtype=dtype)
    
    
    matching_event = np.zeros(inds.size, dtype=dtype)
    
    mask1 = membership_sorted[inds] == 1
    inds1 = inds[mask1]
    n1 = np.sum(mask1)
    matching_event[:n1]['index1'] = spike_index_sorted[inds1]
    matching_event[:n1]['index2'] = spike_index_sorted[inds1+1]
    matching_event[:n1]['delta_frame'] = times_concat_sorted[inds1+1] - times_concat_sorted[inds1]

    mask2 = membership_sorted[inds] == 2
    inds2 = inds[mask2]
    n2 = np.sum(mask2)
    matching_event[n1:]['index1'] = spike_index_sorted[inds2+1]
    matching_event[n1:]['index2'] = spike_index_sorted[inds2]
    matching_event[n1:]['delta_frame'] = times_concat_sorted[inds2] - times_concat_sorted[inds2+1]
    
    order = np.argsort(matching_event['index1'])
    matching_event = matching_event[order]
    
    return matching_event

def make_collision_events(sorting, delta):
    """
    Similar to count_matching_events but get index instead of counting.
    Used for collision detection

    Parameters
    ----------
    sorting: SortingExtractor
        The sorting extractor object for counting collision events
        
    delta: int
        Number of frames for considering collision events

    Returns
    -------
    collision_events: numpy array
            dtype =  [('index1', 'int64'), ('unit_id1', 'int64'),
                ('index2', 'int64'), ('unit_id2', 'int64'),
                ('delta', 'int64')]
        1d of all collision

    """
    dtype = [
            ('index1', 'int64'), ('unit_id1', 'int64'),
            ('index2', 'int64'), ('unit_id2', 'int64'),
            ('delta_frame', 'int64')
        ]
    
    unit_ids = np.array(sorting.get_unit_ids())
    
    collision_events = []
    for i, u1 in enumerate(unit_ids):
        times1 = sorting.get_unit_spike_train(u1)
        
        for u2 in unit_ids[i+1:]:
            times2 = sorting.get_unit_spike_train(u2)
                
            matching_event = make_matching_events(times1, times2, delta)
            ce = np.zeros(matching_event.size, dtype=dtype)
            ce['index1'] = matching_event['index1']
            ce['unit_id1'] = u1
            ce['index2'] = matching_event['index2']
            ce['unit_id2'] = u2
            ce['delta_frame'] = matching_event['delta_frame']
            
            collision_events.append(ce)
    
    collision_events = np.concatenate(collision_events)
    
    return collision_events



import numpy as np
import spikeextractors as se
from pathlib import Path
import json
import os
from .basecomparison import BaseComparison
from .symmetricsortingcomparison import SymmetricSortingComparison
from .comparisontools import compare_spike_trains

import networkx as nx


class MultiSortingComparison(BaseComparison):
    def __init__(self, sorting_list, name_list=None, delta_time=0.4, sampling_frequency=None,
                 match_score=0.5, chance_score=0.1, n_jobs=-1, spiketrain_mode='union', verbose=False,
                 do_matching=True):
        BaseComparison.__init__(self, sorting_list, name_list=name_list,
                                delta_time=delta_time, sampling_frequency=sampling_frequency,
                                match_score=match_score, chance_score=chance_score,
                                n_jobs=n_jobs, verbose=verbose)
        self._spiketrain_mode = spiketrain_mode
        self.clean_graph = None
        if do_matching:
            self._do_comparison()
            self._do_graph()
            self._remove_duplicate_edges()
            self._do_agreement()

    def get_agreement_sorting(self, minimum_agreement_count=1, minimum_agreement_count_only=False):
        '''
        Returns AgreementSortingExtractor with units with a 'minimum_matching' agreement.

        Parameters
        ----------
        minimum_agreement_count: int
            Minimum number of matches among sorters to include a unit.
        minimum_agreement_count_only: bool
            If True, only units with agreement == 'minimum_matching' are included.
            If False, units with an agreement >= 'minimum_matching' are included

        Returns
        -------
        agreement_sorting: AgreementSortingExtractor
            The output AgreementSortingExtractor
        '''
        assert minimum_agreement_count > 0, "'minimum_agreement_count' should be greater than 0"
        sorting = AgreementSortingExtractor(self, min_agreement_count=minimum_agreement_count,
                                            min_agreement_count_only=minimum_agreement_count_only)
        sorting.set_sampling_frequency(self.sampling_frequency)
        return sorting

    def compute_subgraphs(self):
        '''
        Computes subgraphs of connected components.

        Returns
        -------
        sg_sorter_names: list
            List of sorter names for each node in the connected component subrgaph
        sg_units: list
            List of unit ids for each node in the connected component subrgaph
        '''
        if self.clean_graph is not None:
            g = self.clean_graph
        else:
            g = self.graph
        subgraphs = (g.subgraph(c).copy() for c in nx.connected_components(g))
        sg_sorter_names = []
        sg_units = []
        for i, sg in enumerate(subgraphs):
            sorter_names = []
            sorter_units = []
            for node in sg.nodes:
                sorter_names.append(node.split('_')[0])
                sorter_units.append(int(node.split('_')[1]))
            sg_sorter_names.append(sorter_names)
            sg_units.append(sorter_units)
        return sg_sorter_names, sg_units

    def dump(self, save_folder):
        save_folder = Path(save_folder)
        save_folder.mkdir(parents=True, exist_ok=True)
        filename = str(save_folder / 'multicomparison.gpickle')
        nx.write_gpickle(self.graph, filename)
        kwargs = {'delta_time': self.delta_time, 'sampling_frequency': self.sampling_frequency,
                  'match_score': self.match_score, 'chance_score': self.chance_score,
                  'n_jobs': self._n_jobs, 'verbose': self._verbose}
        with (save_folder / 'kwargs.json').open('w') as f:
            json.dump(kwargs, f)
        sortings = {}
        for (name, sort) in zip(self.name_list, self.sorting_list):
            if sort.check_if_dumpable():
                sortings[name] = sort.make_serialized_dict()
            else:
                print(f'Skipping {name} because it is not dumpable')
        with (save_folder / 'sortings.json').open('w') as f:
            json.dump(sortings, f)

    @staticmethod
    def load_multicomparison(folder_path):
        folder_path = Path(folder_path)
        with (folder_path / 'kwargs.json').open() as f:
            kwargs = json.load(f)
        with (folder_path / 'sortings.json').open() as f:
            sortings = json.load(f)
        name_list = sortings.keys()
        sorting_list = [se.load_extractor_from_dict(v) for v in sortings.values()]
        mcmp = MultiSortingComparison(sorting_list=sorting_list, name_list=list(name_list), do_matching=False, **kwargs)
        mcmp.graph = nx.read_gpickle(str(folder_path / 'multicomparison.gpickle'))
        # do step 3 and 4
        mcmp._remove_duplicate_edges()
        mcmp._do_agreement()
        return mcmp

    def _do_comparison(self,):
        # do pairwise matching
        if self._verbose:
            print('Multicomaprison step 1: pairwise comparison')

        self.comparisons = []
        for i in range(len(self.sorting_list)):
            for j in range(i + 1, len(self.sorting_list)):
                if self._verbose:
                    print("  Comparing: ", self.name_list[i], " and ", self.name_list[j])
                comp = SymmetricSortingComparison(self.sorting_list[i], self.sorting_list[j],
                                                  sorting1_name=self.name_list[i],
                                                  sorting2_name=self.name_list[j],
                                                  delta_time=self.delta_time,
                                                  sampling_frequency=self.sampling_frequency,
                                                  match_score=self.match_score,
                                                  n_jobs=self._n_jobs,
                                                  verbose=False)
                self.comparisons.append(comp)

    def _do_graph(self):
        if self._verbose:
            print('Multicomaprison step 2: make graph')

        self.graph = nx.Graph()
        # nodes
        for i, sorting in enumerate(self.sorting_list):
            sorter_name = self.name_list[i]
            for unit in sorting.get_unit_ids():
                node_name = str(sorter_name) + '_' + str(unit)
                self.graph.add_node(node_name)
        # edges
        for comp in self.comparisons:
            for u1 in comp.sorting1.get_unit_ids():
                u2 = comp.hungarian_match_12[u1]
                if u2 != -1:
                    node1_name = str(comp.sorting1_name) + '_' + str(u1)
                    node2_name = str(comp.sorting2_name) + '_' + str(u2)
                    score = comp.agreement_scores.loc[u1, u2]
                    self.graph.add_edge(node1_name, node2_name, weight=score)

        # the graph is symmetrical
        self.graph = self.graph.to_undirected()

    def _do_agreement(self):
        # extract agreement from graph
        if self._verbose:
            print('Multicomaprison step 4: extract agreement from graph')

        self._new_units = {}
        self._spiketrains = []
        unit_id = 0

        # Remove duplicate
        subgraphs = (self.clean_graph.subgraph(c).copy() for c in nx.connected_components(self.clean_graph))
        for i, sg in enumerate(subgraphs):
            edges = list(sg.edges(data=True))
            if len(edges) > 0:
                avg_agr = np.mean([d['weight'] for u, v, d in edges])
            else:
                avg_agr = 0
            sorter_unit_ids = {}
            for node in sg.nodes:
                sorter_name = node.split('_')[0]
                sorter_unit = int(node.split('_')[1])
                sorter_unit_ids[sorter_name] = sorter_unit
            self._new_units[unit_id] = {'avg_agreement': avg_agr, 'sorter_unit_ids': sorter_unit_ids,
                                        'agreement_number': len(sg.nodes)}
            # Append correct spike train
            if len(sorter_unit_ids.keys()) == 1:
                self._spiketrains.append(self.sorting_list[self.name_list.index(
                    list(sorter_unit_ids.keys())[0])].get_unit_spike_train(list(sorter_unit_ids.values())[0]))
            else:
                max_edge = edges[int(np.argmax([d['weight'] for u, v, d in edges]))]
                node1, node2, weight = max_edge
                sorter1, unit1 = node1.split('_')
                sorter2, unit2 = node2.split('_')
                unit1 = int(unit1)
                unit2 = int(unit2)
                sp1 = self.sorting_list[self.name_list.index(sorter1)].get_unit_spike_train(unit1)
                sp2 = self.sorting_list[self.name_list.index(sorter2)].get_unit_spike_train(unit2)

                if self._spiketrain_mode == 'union':
                    lab1, lab2 = compare_spike_trains(sp1, sp2)
                    # add FP to spike train 1 (FP are the only spikes outside the union)
                    fp_idx2 = np.where(np.array(lab2) == 'FP')[0]
                    sp_union = np.sort(np.concatenate((sp1, sp2[fp_idx2])))
                    self._spiketrains.append(list(sp_union))
                elif self._spiketrain_mode == 'intersection':
                    lab1, lab2 = compare_spike_trains(sp1, sp2)
                    # TP are the spikes in the intersection
                    tp_idx1 = np.where(np.array(lab1) == 'TP')[0]
                    sp_tp1 = list(np.array(sp1)[tp_idx1])
                    self._spiketrains.append(sp_tp1)
            unit_id += 1

    def _do_agreement_matrix(self, minimum_agreement=1):
        sorted_name_list = sorted(self.name_list)
        sorting_agr = AgreementSortingExtractor(self, minimum_agreement)
        unit_ids = sorting_agr.get_unit_ids()
        agreement_matrix = np.zeros((len(unit_ids), len(sorted_name_list)))

        for u_i, unit in enumerate(unit_ids):
            for sort_name, sorter in enumerate(sorted_name_list):
                if sorter in sorting_agr.get_unit_property(unit, 'sorter_unit_ids').keys():
                    assigned_unit = sorting_agr.get_unit_property(unit, 'sorter_unit_ids')[sorter]
                else:
                    assigned_unit = -1
                if assigned_unit == -1:
                    agreement_matrix[u_i, sort_name] = np.nan
                else:
                    agreement_matrix[u_i, sort_name] = sorting_agr.get_unit_property(unit, 'avg_agreement')
        return agreement_matrix

    def _remove_duplicate_edges(self):
        if self._verbose:
            print('Multicomaprison step 3: clean graph')
        clean_graph = self.graph.copy()
        subgraphs = (clean_graph.subgraph(c).copy() for c in nx.connected_components(clean_graph))
        removed_nodes = 0
        for i, sg in enumerate(subgraphs):
            sorter_names = []
            for node in sg.nodes:
                sorter_names.append(node.split('_')[0])
            sorters, counts = np.unique(sorter_names, return_counts=True)

            if np.any(counts > 1):
                for sorter in sorters[counts > 1]:
                    nodes_duplicate = [n for n in sg.nodes if sorter in n]
                    # get edges
                    edges_duplicates = []
                    weights_duplicates = []
                    for n in nodes_duplicate:
                        edges = sg.edges(n, data=True)
                        for e in edges:
                            edges_duplicates.append(e)
                            weights_duplicates.append(e[2]['weight'])
                    # remove edges
                    edges_to_remove = len(nodes_duplicate) - 1
                    remove_idxs = np.argsort(weights_duplicates)[:edges_to_remove]
                    for idx in remove_idxs:
                        if self._verbose:
                            print('Removed edge', edges_duplicates[idx])
                        clean_graph.remove_edge(edges_duplicates[idx][0], edges_duplicates[idx][1])
                        sg.remove_edge(edges_duplicates[idx][0], edges_duplicates[idx][1])
                        if edges_duplicates[idx][0] in nodes_duplicate:
                            sg.remove_node(edges_duplicates[idx][0])
                        else:
                            sg.remove_node(edges_duplicates[idx][1])
                        removed_nodes += 1
        if self._verbose:
            print(f'Removed {removed_nodes} duplicate nodes')
        self.clean_graph = clean_graph


class AgreementSortingExtractor(se.SortingExtractor):

    def __init__(self, multisortingcomparison, min_agreement_count=1, min_agreement_count_only=False):
        se.SortingExtractor.__init__(self)
        self._msc = multisortingcomparison
        self.is_dumpable = False

        if min_agreement_count_only:
            self._unit_ids = list(u for u in self._msc._new_units.keys()
                                  if self._msc._new_units[u]['agreement_number'] == min_agreement_count)
        else:
            self._unit_ids = list(u for u in self._msc._new_units.keys()
                                  if self._msc._new_units[u]['agreement_number'] >= min_agreement_count)

        for unit in self._unit_ids:
            self.set_unit_property(unit_id=unit, property_name='agreement_number',
                                   value=self._msc._new_units[unit]['agreement_number'])
            self.set_unit_property(unit_id=unit, property_name='avg_agreement',
                                   value=self._msc._new_units[unit]['avg_agreement'])
            self.set_unit_property(unit_id=unit, property_name='sorter_unit_ids',
                                   value=self._msc._new_units[unit]['sorter_unit_ids'])

    def get_unit_ids(self, unit_ids=None):
        if unit_ids is None:
            return self._unit_ids
        else:
            return self._unit_ids[unit_ids]

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if unit_id not in self.get_unit_ids():
            raise Exception("Unit id is invalid")
        return np.array(self._msc._spiketrains[list(self._msc._new_units.keys()).index(unit_id)])


def compare_multiple_sorters(sorting_list, name_list=None, delta_time=0.4, match_score=0.5, chance_score=0.1,
                             n_jobs=-1, spiketrain_mode='union', sampling_frequency=None, verbose=False):
    '''
    Compares multiple spike sorter outputs.

    - Pair-wise comparisons are made
    - An agreement graph is built based on the agreement score

    It allows to return a consensus-based sorting extractor with the `get_agreement_sorting()` method.

    Parameters
    ----------
    sorting_list: list
        List of sorting extractor objects to be compared
    name_list: list
        List of spike sorter names. If not given, sorters are named as 'sorter0', 'sorter1', 'sorter2', etc.
    delta_time: float
        Number of ms to consider coincident spikes (default 0.4 ms)
    match_score: float
        Minimum agreement score to match units (default 0.5)
    chance_score: float
        Minimum agreement score to for a possible match (default 0.1)
    n_jobs: int
       Number of cores to use in parallel. Uses all availible if -1
    spiketrain_mode: str
        Mode to extract agreement spike trains:
            - 'union': spike trains are the union between the spike trains of the best matching two sorters
            - 'intersection': spike trains are the intersection between the spike trains of the
               best matching two sorters
    sampling_frequency: float
        Sampling frequency (used if information is not in the sorting extractors)
    verbose: bool
        if True, output is verbose

    Returns
    -------
    multi_sorting_comparison: MultiSortingComparison
        MultiSortingComparison object with the multiple sorter comparison
    '''
    return MultiSortingComparison(sorting_list=sorting_list, name_list=name_list, delta_time=delta_time,
                                  match_score=match_score, chance_score=chance_score, n_jobs=n_jobs,
                                  spiketrain_mode=spiketrain_mode, sampling_frequency=sampling_frequency,
                                  verbose=verbose)




from .version import version as __version__

from .comparisontools import (count_matching_events, compute_agreement_score, count_match_spikes,
        make_agreement_scores, make_possible_match, make_best_match, make_hungarian_match, 
        do_score_labels, compare_spike_trains, do_confusion_matrix, do_count_score, compute_performence)
from .symmetricsortingcomparison import compare_two_sorters, SymmetricSortingComparison
from .groundtruthcomparison import compare_sorter_to_ground_truth, GroundTruthComparison
from .multisortingcomparison import compare_multiple_sorters, MultiSortingComparison

from .collisioncomparison import CollisionGTComparison

from .groundtruthstudy import GroundTruthStudy



import pandas as pd
import numpy as np
from .groundtruthcomparison import GroundTruthComparison
#~ from .comparisontools import (do_score_labels, make_possible_match,
                              #~ make_best_match, make_hungarian_match, do_confusion_matrix, do_count_score,
                              #~ compute_performence)
from .comparisontools import make_collision_events


class CollisionGTComparison(GroundTruthComparison):
    """
    This class is an extention of GroundTruthComparison by focusing
    to benchmark spike in collision
    
    
    collision_lag: float 
        Collision lag in ms.
    
    """

    def __init__(self, gt_sorting, tested_sorting, collision_lag=2.0, **kwargs):
        
        # Force compute labels
        kwargs['compute_labels'] = True
        
        GroundTruthComparison.__init__(self, gt_sorting, tested_sorting, **kwargs)
        
        self.collision_lag = collision_lag
        self.detect_gt_collision()
        
    def detect_gt_collision(self):
        delta = int(self.collision_lag / 1000 * self.sampling_frequency)
        self.collision_events = make_collision_events(self.sorting1, delta)
    
    def get_label_for_collision(self, gt_unit_id1, gt_unit_id2):
        if gt_unit_id1 > gt_unit_id2:
            gt_unit_id1, gt_unit_id2 = gt_unit_id2, gt_unit_id1
            reversed = True
        else:
            reversed = False
        
        # events
        mask = (self.collision_events['unit_id1'] == gt_unit_id1) & (self.collision_events['unit_id2'] == gt_unit_id2)
        event = self.collision_events[mask]
        
        score_label1 = self._labels_st1[gt_unit_id1][event['index1']]
        score_label2 = self._labels_st1[gt_unit_id2][event['index2']]
        delta = event['delta_frame']
        
        if reversed:
            score_label1, score_label2 = score_label2, score_label1
            delta = -delta
        
        return score_label1, score_label2, delta
    
    def get_label_count_per_collision_bins(self, gt_unit_id1, gt_unit_id2, nbins=11):
        d = int(self.collision_lag / 1000 * self.sampling_frequency)
        bins = np.arange(-d, d, d/10.)
        
        score_label1, score_label2, delta = self.get_label_for_collision(gt_unit_id1, gt_unit_id2)
        
        tp_count1 = np.zeros(bins.size-1)
        fn_count1 = np.zeros(bins.size-1)
        tp_count2 = np.zeros(bins.size-1)
        fn_count2 = np.zeros(bins.size-1)
        
        for i in range(tp_count1.size):
            l0, l1 = bins[i], bins[i+1]
            mask = (delta>=l0) & (delta<l1)
            
            tp_count1[i] = np.sum(score_label1[mask] == 'TP')
            fn_count1[i] = np.sum(score_label1[mask] == 'FN')
            tp_count2[i] = np.sum(score_label2[mask] == 'TP')
            fn_count2[i] = np.sum(score_label2[mask] == 'FN')
        
        # inverse for unit_id2
        tp_count2 = tp_count2[::-1]
        fn_count2 = fn_count2[::-1]

        return bins, tp_count1, fn_count1, tp_count2, fn_count2

        
    



import pandas as pd
import numpy as np
from .basecomparison import BaseTwoSorterComparison
from .comparisontools import (do_score_labels, make_possible_match,
                              make_best_match, make_hungarian_match, do_confusion_matrix, do_count_score,
                              compute_performence)


class GroundTruthComparison(BaseTwoSorterComparison):
    """
    Class to compare a sorter to ground truth (GT)
    
    This class can:
      * compute a "macth between gt_sorting and tested_sorting
      * compte th score label (TP, FN, CL, FP) for each spike
      * count by unit of GT the total of each (TP, FN, CL, FP) into a Dataframe 
        GroundTruthComparison.count
      * compute the confusion matrix .get_confusion_matrix()
      * compute some performance metric with several strategy based on 
        the count score by unit
      * count well detected units
      * count false positve detected units
      * count redundant units
      * count overmerged units
      * summary all this
    """

    def __init__(self, gt_sorting, tested_sorting, gt_name=None, tested_name=None,
                 delta_time=0.4, sampling_frequency=None, match_score=0.5, well_detected_score=0.8,
                 redundant_score=0.2, overmerged_score=0.2, chance_score=0.1, exhaustive_gt=False, n_jobs=-1,
                 match_mode='hungarian', compute_labels=False, compute_misclassifications=False, verbose=False):

        if gt_name is None:
            gt_name = 'ground truth'
        if tested_name is None:
            tested_name = 'tested'
        BaseTwoSorterComparison.__init__(self, gt_sorting, tested_sorting, sorting1_name=gt_name,
                                         sorting2_name=tested_name, delta_time=delta_time,
                                         sampling_frequency=sampling_frequency, match_score=match_score,
                                         chance_score=chance_score, n_jobs=n_jobs,
                                         verbose=verbose)
        self.exhaustive_gt = exhaustive_gt

        self._compute_misclassifications = compute_misclassifications
        self.redundant_score = redundant_score
        self.overmerged_score = overmerged_score
        self.well_detected_score = well_detected_score

        assert match_mode in ['hungarian', 'best']
        self.match_mode = match_mode
        self._compute_labels = compute_labels

        self._do_count()

        self._labels_st1 = None
        self._labels_st2 = None
        if self._compute_labels:
            self._do_score_labels()

        # confusion matrix is compute on demand
        self._confusion_matrix = None

    def get_labels1(self, unit_id):
        if self._labels_st1 is None:
            self._do_score_labels()

        if unit_id in self.sorting1.get_unit_ids():
            return self._labels_st1[unit_id]
        else:
            raise Exception("Unit_id is not a valid unit")

    def get_labels2(self, unit_id):
        if self._labels_st1 is None:
            self._do_score_labels()

        if unit_id in self.sorting2.get_unit_ids():
            return self._labels_st2[unit_id]
        else:
            raise Exception("Unit_id is not a valid unit")

    def _do_matching(self):
        if self._verbose:
            print("Matching...")

        self.possible_match_12, self.possible_match_21 = make_possible_match(self.agreement_scores, self.chance_score)
        self.best_match_12, self.best_match_21 = make_best_match(self.agreement_scores, self.chance_score)
        self.hungarian_match_12, self.hungarian_match_21 = make_hungarian_match(self.agreement_scores,
                                                                                self.match_score)

    def _do_count(self):
        """
        Do raw count into a dataframe.
        
        Internally use hugarian match or best match.
        
        """
        if self.match_mode == 'hungarian':
            match_12 = self.hungarian_match_12
        elif self.match_mode == 'best':
            match_12 = self.best_match_12

        self.count_score = do_count_score(self.event_counts1, self.event_counts2,
                                          match_12, self.match_event_count)

    def _do_confusion_matrix(self):
        if self._verbose:
            print("Computing confusion matrix...")

        if self.match_mode == 'hungarian':
            match_12 = self.hungarian_match_12
        elif self.match_mode == 'best':
            match_12 = self.best_match_12

        self._confusion_matrix = do_confusion_matrix(self.event_counts1, self.event_counts2, match_12,
                                                     self.match_event_count)

    def get_confusion_matrix(self):
        """
        Computes the confusion matrix.

        Returns
        -------
        confusion_matrix: pandas.DataFrame
            The confusion matrix
        """
        if self._confusion_matrix is None:
            self._do_confusion_matrix()
        return self._confusion_matrix

    def _do_score_labels(self):
        assert self.match_mode == 'hungarian', \
            'Labels (TP, FP, FN) can be computed only with hungarian match'

        if self._verbose:
            print("Adding labels...")

        self._labels_st1, self._labels_st2 = do_score_labels(self.sorting1, self.sorting2,
                                                             self.delta_frames, self.hungarian_match_12,
                                                             self._compute_misclassifications)

    def get_performance(self, method='by_unit', output='pandas'):
        """
        Get performance rate with several method:
          * 'raw_count' : just render the raw count table
          * 'by_unit' : render perf as rate unit by unit of the GT
          * 'pooled_with_average' : compute rate unit by unit and average

        Parameters
        ----------
        method: str
            'by_unit',  or 'pooled_with_average'
        output: str
            'pandas' or 'dict'

        Returns
        -------
        perf: pandas dataframe/series (or dict)
            dataframe/series (based on 'output') with performance entries
        """
        possibles = ('raw_count', 'by_unit', 'pooled_with_average')
        if method not in possibles:
            raise Exception("'method' can be " + ' or '.join(possibles))

        if method == 'raw_count':
            perf = self.count_score

        elif method == 'by_unit':
            perf = compute_performence(self.count_score)

        elif method == 'pooled_with_average':
            perf = self.get_performance(method='by_unit').mean(axis=0)

        if output == 'dict' and isinstance(perf, pd.Series):
            perf = perf.to_dict()

        return perf

    def print_performance(self, method='pooled_with_average'):
        """
        Print performance with the selected method
        """

        template_txt_performance = _template_txt_performance

        if method == 'by_unit':
            perf = self.get_performance(method=method, output='pandas')
            perf = perf * 100
            # ~ print(perf)
            d = {k: perf[k].tolist() for k in perf.columns}
            txt = template_txt_performance.format(method=method, **d)
            print(txt)

        elif method == 'pooled_with_average':
            perf = self.get_performance(method=method, output='pandas')
            perf = perf * 100
            txt = template_txt_performance.format(method=method, **perf.to_dict())
            print(txt)

    def print_summary(self, well_detected_score=None, redundant_score=None, overmerged_score=None):
        """
        Print a global performance summary that depend on the context:
          * exhaustive= True/False
          * how many gt units (one or several)
        
        This summary mix several performance metrics.
        """
        txt = _template_summary_part1

        d = dict(
            num_gt=len(self.unit1_ids),
            num_tested=len(self.unit2_ids),
            num_well_detected=self.count_well_detected_units(well_detected_score),
            num_redundant=self.count_redundant_units(redundant_score),
            num_overmerged=self.count_overmerged_units(overmerged_score),
        )

        if self.exhaustive_gt:
            txt = txt + _template_summary_part2
            d['num_false_positive_units'] = self.count_false_positive_units()
            d['num_bad'] = self.count_bad_units()

        txt = txt.format(**d)

        print(txt)

    def get_well_detected_units(self, well_detected_score=None):
        """
        Return units list of "well detected units" from tested_sorting.

        "well detected units" ara defined as units in tested that
        are well matched to GT units.

        Parameters
        ----------
        well_detected_score: float (default 0.8)
            The agreement score above which tested units
            are counted as "well detected".
        """
        if well_detected_score is not None:
            self.well_detected_score = well_detected_score

        matched_units2 = self.hungarian_match_12
        well_detected_ids = []
        for u2 in self.unit2_ids:
            if u2 in list(matched_units2.values):
                u1 = self.hungarian_match_21[u2]
                score = self.agreement_scores.at[u1, u2]
                if score >= self.well_detected_score:
                    well_detected_ids.append(u2)

        return well_detected_ids

    def count_well_detected_units(self, well_detected_score):
        """
        Count how many well detected units.
        Kargs are the same as get_well_detected_units.
        """
        return len(self.get_well_detected_units(well_detected_score=well_detected_score))

    def get_false_positive_units(self, redundant_score=None):
        """
        Return units list of "false positive units" from tested_sorting.
        
        "false positive units" ara defined as units in tested that
        are not matched at all in GT units.
        
        Need exhaustive_gt=True

        Parameters
        ----------
        redundant_score: float (default 0.2)
            The agreement score below which tested units
            are counted as "false positive"" (and not "redundant").
        """
        assert self.exhaustive_gt, 'false_positive_units list is valid only if exhaustive_gt=True'

        if redundant_score is not None:
            self.redundant_score = redundant_score

        matched_units2 = list(self.hungarian_match_12.values)
        false_positive_ids = []
        for u2 in self.unit2_ids:
            if u2 not in matched_units2:
                if self.best_match_21[u2] == -1:
                    false_positive_ids.append(u2)
                else:
                    u1 = self.best_match_21[u2]
                    score = self.agreement_scores.at[u1, u2]
                    if score < self.redundant_score:
                        false_positive_ids.append(u2)

        return false_positive_ids

    def count_false_positive_units(self, redundant_score=None):
        """
        See get_false_positive_units().
        """
        return len(self.get_false_positive_units(redundant_score))

    def get_redundant_units(self, redundant_score=None):
        """
        Return "redundant units"
        
        
        "redundant units" are defined as units in tested
        that match a GT units with a big agreement score
        but it is not the best match.
        In other world units in GT that detected twice or more.
        
        Parameters
        ----------
        redundant_score=None: float (default 0.2)
            The agreement score above which tested units
            are counted as "redundant" (and not "false positive" ).
        """
        assert self.exhaustive_gt, 'redundant_units list is valid only if exhaustive_gt=True'

        if redundant_score is not None:
            self.redundant_score = redundant_score
        matched_units2 = list(self.hungarian_match_12.values)
        redundant_ids = []
        for u2 in self.unit2_ids:
            if u2 not in matched_units2 and self.best_match_21[u2] != -1:
                u1 = self.best_match_21[u2]
                if u2 != self.best_match_12[u1]:
                    score = self.agreement_scores.at[u1, u2]
                    if score >= self.redundant_score:
                        redundant_ids.append(u2)

        return redundant_ids

    def count_redundant_units(self, redundant_score=None):
        """
        See get_redundant_units().
        """
        return len(self.get_redundant_units(redundant_score=redundant_score))

    def get_overmerged_units(self, overmerged_score=None):
        """
        Return "overmerged units"


        "overmerged units" are defined as units in tested
        that match more than one GT unit with an agreement score larger than overmerged_score.

        Parameters
        ----------
        overmerged_score: float (default 0.4)
            Tested units with 2 or more agrement scores above 'overmerged_score'
            are counted as "overmerged".
        """
        assert self.exhaustive_gt, 'overmerged_units list is valid only if exhaustive_gt=True'

        if overmerged_score is not None:
            self.overmerged_score = overmerged_score

        overmerged_ids = []
        for u2 in self.unit2_ids:
            scores = self.agreement_scores.loc[:, u2]
            if len(np.where(scores > self.overmerged_score)[0]) > 1:
                overmerged_ids.append(u2)

        return overmerged_ids

    def count_overmerged_units(self, overmerged_score=None):
        """
        See get_overmerged_units().
        """
        return len(self.get_overmerged_units(overmerged_score=overmerged_score))

    def get_bad_units(self):
        """
        Return units list of "bad units".
        
        "bad units" are defined as units in tested that are not
        in the best match list of GT units.
        
        So it is the union of "false positive units" + "redundant units".
        
        Need exhaustive_gt=True
        """
        assert self.exhaustive_gt, 'bad_units list is valid only if exhaustive_gt=True'
        matched_units2 = list(self.hungarian_match_12.values)
        bad_ids = []
        for u2 in self.unit2_ids:
            if u2 not in matched_units2:
                bad_ids.append(u2)
        return bad_ids

    def count_bad_units(self):
        """
        See get_bad_units
        """
        return len(self.get_bad_units())


# usefull also for gathercomparison


_template_txt_performance = """PERFORMANCE ({method})
-----------
ACCURACY: {accuracy}
RECALL: {recall}
PRECISION: {precision}
FALSE DISCOVERY RATE: {false_discovery_rate}
MISS RATE: {miss_rate}
"""

_template_summary_part1 = """SUMMARY
-------
GT num_units: {num_gt}
TESTED num_units: {num_tested}
num_well_detected: {num_well_detected} 
num_redundant: {num_redundant}
num_overmerged: {num_overmerged}
"""

_template_summary_part2 = """num_false_positive_units {num_false_positive_units}
num_bad: {num_bad}
"""


def compare_sorter_to_ground_truth(gt_sorting, tested_sorting, gt_name=None, tested_name=None,
                                   delta_time=0.4, sampling_frequency=None, match_score=0.5, chance_score=0.1,
                                   well_detected_score=0.8, redundant_score=0.2, overmerged_score=0.2,
                                   exhaustive_gt=True, match_mode='hungarian', n_jobs=-1, compute_labels=False,
                                   compute_misclassifications=False, verbose=False):
    '''
    Compares a sorter to a ground truth.

    - Spike trains are matched based on their agreement scores
    - Individual spikes are labelled as true positives (TP), false negatives (FN),
      false positives 1 (FP), misclassifications (CL)

    It also allows to compute_performance and confusion matrix.

    Parameters
    ----------
    gt_sorting: SortingExtractor
        The first sorting for the comparison
    tested_sorting: SortingExtractor
        The second sorting for the comparison
    gt_name: str
        The name of sorter 1
    tested_name: : str
        The name of sorter 2
    delta_time: float
        Number of ms to consider coincident spikes (default 0.4 ms)
    sampling_frequency: float
        Optional sampling frequency in Hz when not included in sorting
    match_score: float
        Minimum agreement score to match units (default 0.5)
    chance_score: float
        Minimum agreement score to for a possible match (default 0.1)
    redundant_score: float
        Agreement score above which units are redundant (default 0.2)
    overmerged_score: float
        Agreement score above which units can be overmerged (default 0.2)
    well_detected_score: float
        Agreement score above which units are well detected (default 0.8)
    exhaustive_gt: bool (default True)
        Tell if the ground true is "exhaustive" or not. In other world if the
        GT have all possible units. It allows more performance measurement.
        For instance, MEArec simulated dataset have exhaustive_gt=True
    match_mode: 'hungarian', or 'best'
        What is match used for counting : 'hugarian' or 'best match'.
    n_jobs: int
        Number of cores to use in parallel. Uses all available if -1
    compute_labels: bool
        If True, labels are computed at instantiation (default False)
    compute_misclassifications: bool
        If True, misclassifications are computed at instantiation (default False)
    verbose: bool
        If True, output is verbose
    Returns
    -------
    sorting_comparison: SortingComparison
        The SortingComparison object
    '''
    return GroundTruthComparison(gt_sorting=gt_sorting, tested_sorting=tested_sorting, gt_name=gt_name,
                                 tested_name=tested_name, delta_time=delta_time, sampling_frequency=sampling_frequency,
                                 match_score=match_score, redundant_score=redundant_score, chance_score=chance_score,
                                 well_detected_score=well_detected_score, overmerged_score=overmerged_score,
                                 exhaustive_gt=exhaustive_gt, n_jobs=n_jobs,
                                 match_mode=match_mode, compute_labels=compute_labels,
                                 compute_misclassifications=compute_misclassifications, verbose=verbose)



import numpy as np
import pandas as pd
from .comparisontools import (do_count_event, make_match_count_matrix, make_agreement_scores_from_count)


class BaseComparison:
    """
    Base class for all comparison classes:
       * GroundTruthComparison
       * MultiSortingComparison
       * SymmetricSortingComparison
    
    Mainly deals with:
      * sampling_frequency
      * sorting names
      * delta_time to delta_frames
    """

    def __init__(self, sorting_list, name_list=None, delta_time=0.4, sampling_frequency=None,
                 match_score=0.5, chance_score=0.1, n_jobs=-1, verbose=False):

        self.sorting_list = sorting_list
        if name_list is None:
            name_list = ['sorting{}'.format(i + 1) for i in range(len(sorting_list))]
        self.name_list = name_list
        if np.any(['_' in name for name in name_list]):
            raise ValueError("Sorter names in 'name_list' cannot contain '_'")

        if sampling_frequency is None:
            # take sampling frequency from sorting list and test that they are equivalent.
            sampling_freqs_not_none = np.array([s.get_sampling_frequency() for s in self.sorting_list
                                                if s.get_sampling_frequency() is not None], dtype='float64')
            assert len(sampling_freqs_not_none) > 0, ("Sampling frequency information "
                                                      "not found in sorting. Pass it with the 'sampling_frequency' "
                                                      "argument")

            # Some sorter round the sampling freq lets emit a warning
            sf0 = sampling_freqs_not_none[0]
            if not np.all([sf == sf0 for sf in sampling_freqs_not_none]):
                delta_freq_ratio = np.abs(sampling_freqs_not_none - sf0) / sf0
                # tolerence of 0.1%
                assert np.all(delta_freq_ratio < 0.001), "Inconsintent sampling frequency among sorting list"

            sampling_frequency = sampling_freqs_not_none[0]

        self.sampling_frequency = sampling_frequency
        self.delta_time = delta_time
        self.delta_frames = int(self.delta_time / 1000 * self.sampling_frequency)
        self.match_score = match_score
        self.chance_score = chance_score
        self._n_jobs = n_jobs
        self._verbose = verbose


class BaseTwoSorterComparison(BaseComparison):
    """
    Base class shared by SortingComparison and GroundTruthComparison
    """

    def __init__(self, sorting1, sorting2, sorting1_name=None, sorting2_name=None,
                 delta_time=0.4, sampling_frequency=None, match_score=0.5,
                 chance_score=0.1, n_jobs=1, verbose=False):

        sorting_list = [sorting1, sorting2]
        if sorting1_name is None:
            sorting1_name = 'sorting1'
        if sorting2_name is None:
            sorting2_name = 'sorting2'
        name_list = [sorting1_name, sorting2_name]

        BaseComparison.__init__(self, sorting_list, name_list=name_list, delta_time=delta_time,
                                sampling_frequency=sampling_frequency, match_score=match_score,
                                chance_score=chance_score, verbose=verbose, n_jobs=n_jobs)

        self.unit1_ids = self.sorting1.get_unit_ids()
        self.unit2_ids = self.sorting2.get_unit_ids()

        self._do_agreement()
        self._do_matching()

    @property
    def sorting1(self):
        return self.sorting_list[0]

    @property
    def sorting2(self):
        return self.sorting_list[1]

    @property
    def sorting1_name(self):
        return self.name_list[0]

    @property
    def sorting2_name(self):
        return self.name_list[1]

    def _do_agreement(self):
        if self._verbose:
            print('Agreement scores...')

        # common to GroundTruthComparison and SymmetricSortingComparison
        # spike count for each spike train
        self.event_counts1 = do_count_event(self.sorting1)
        self.event_counts2 = do_count_event(self.sorting2)

        # matrix of  event match count for each pair
        self.match_event_count = make_match_count_matrix(self.sorting1, self.sorting2, self.delta_frames,
                                                         n_jobs=self._n_jobs)

        # agreement matrix score for each pair
        self.agreement_scores = make_agreement_scores_from_count(self.match_event_count, self.event_counts1,
                                                                 self.event_counts2)

    def _do_matching(self):
        # must be implemented in subclass
        raise NotImplementedError

    def get_ordered_agreement_scores(self):
        # order rows
        order0 = self.agreement_scores.max(axis=1).argsort()
        scores = self.agreement_scores.iloc[order0.values[::-1], :]

        # order columns
        indexes = np.arange(scores.shape[1])
        order1 = []
        for r in range(scores.shape[0]):
            possible = indexes[~np.in1d(indexes, order1)]
            if possible.size > 0:
                ind = np.argmax(scores.iloc[r, possible].values)
                order1.append(possible[ind])
        remain = indexes[~np.in1d(indexes, order1)]
        order1.extend(remain)
        scores = scores.iloc[:, order1]

        return scores



"""
High level tools to run many ground-truth comparison with
many sorter on many recordings and then collect and aggregate results
in an easy way.

The all mechanism is based on an intrinsic organisation
into a "study_folder" with several subfolder:
  * raw_files : contain a copy in binary format of recordings
  * sorter_folders : contains output of sorters
  * ground_truth : contains a copy of sorting ground  in npz format
  * sortings: contains light copy of all sorting in npz format
  * tables: some table in cvs format
"""

from pathlib import Path
import shutil
import json
import os

import pandas as pd
import spikeextractors as se

from spikeextractors.extraction_tools import read_python

from spikesorters.sorterlist import sorter_dict

from spikesorters import run_sorters, iter_output_folders, iter_sorting_output
from .comparisontools import _perf_keys
from .groundtruthcomparison import compare_sorter_to_ground_truth


def setup_comparison_study(study_folder, gt_dict):
    """
    Based on a dict of (recording, sorting) create the study folder.


    Parameters
    ----------
    study_folder: str
        The study folder.

    gt_dict : a dict of tuple (recording, sorting_gt)
        Dict of tuple that contain recording and sorting ground truth
    """

    study_folder = Path(study_folder)
    assert not study_folder.is_dir(), "'study_folder' already exists. Please remove it"

    study_folder.mkdir(parents=True, exist_ok=True)
    (study_folder / 'raw_files').mkdir(parents=True, exist_ok=True)
    (study_folder / 'ground_truth').mkdir(parents=True, exist_ok=True)
    (study_folder / 'sortings').mkdir(parents=True, exist_ok=True)
    (study_folder / 'sortings' / 'run_log').mkdir(parents=True, exist_ok=True)

    for rec_name, (recording, sorting_gt) in gt_dict.items():
        # write recording as binary format + json + prb
        raw_filename = study_folder / 'raw_files' / (rec_name + '.dat')
        prb_filename = study_folder / 'raw_files' / (rec_name + '.prb')
        json_filename = study_folder / 'raw_files' / (rec_name + '.json')
        num_chan = recording.get_num_channels()
        chunk_size = 2 ** 24 // num_chan
        sr = recording.get_sampling_frequency()

        recording.write_to_binary_dat_format(raw_filename, time_axis=0, dtype='float32', chunk_size=chunk_size)
        recording.save_to_probe_file(prb_filename)
        with open(json_filename, 'w', encoding='utf8') as f:
            info = dict(sample_rate=sr, num_chan=num_chan, dtype='float32', time_axis=0)
            json.dump(info, f, indent=4)

        # write recording sorting_gt as with npz format
        se.NpzSortingExtractor.write_sorting(sorting_gt, study_folder / 'ground_truth' / (rec_name + '.npz'))

    # make an index of recording names
    with open(study_folder / 'names.txt', mode='w', encoding='utf8') as f:
        for rec_name in gt_dict:
            f.write(rec_name + '\n')


def get_rec_names(study_folder):
    """
    Get list of keys of recordings.
    Read from the 'names.txt' file in study folder.

    Parameters
    ----------
    study_folder: str
        The study folder.

    Returns
    ----------

    rec_names: list
        LIst of names.
    """
    with open(study_folder / 'names.txt', mode='r', encoding='utf8') as f:
        rec_names = f.read()[:-1].split('\n')
    return rec_names


def get_one_recording(study_folder, rec_name):
    """
    Get one recording from its name

    Parameters
    ----------
    study_folder: str
        The study folder.
    rec_name: str
        The recording name
    Returns
    ----------

    recording: RecordingExtractor
        The recording.
    
    """
    raw_filename = study_folder / 'raw_files' / (rec_name + '.dat')
    prb_filename = study_folder / 'raw_files' / (rec_name + '.prb')
    json_filename = study_folder / 'raw_files' / (rec_name + '.json')
    with open(json_filename, 'r', encoding='utf8') as f:
        info = json.load(f)
    rec = se.BinDatRecordingExtractor(raw_filename, info['sample_rate'], info['num_chan'],
                                      info['dtype'], time_axis=info['time_axis'])
    load_probe_file_inplace(rec, prb_filename)

    return rec


def get_recordings(study_folder):
    """
    Get ground recording as a dict.

    They are read from the 'raw_files' folder with binary format.

    Parameters
    ----------
    study_folder: str
        The study folder.

    Returns
    ----------

    recording_dict: dict
        Dict of rexording.

    """
    study_folder = Path(study_folder)

    rec_names = get_rec_names(study_folder)
    recording_dict = {}
    for rec_name in rec_names:
        recording_dict[rec_name] = get_one_recording(study_folder, rec_name)

    return recording_dict


def get_ground_truths(study_folder):
    """
    Get ground truth sorting extractor as a dict.

    They are read from the 'ground_truth' folder with npz format.

    Parameters
    ----------
    study_folder: str
        The study folder.

    Returns
    ----------

    ground_truths: dict
        Dict of sorintg_gt.

    """
    study_folder = Path(study_folder)
    rec_names = get_rec_names(study_folder)
    ground_truths = {}
    for rec_name in rec_names:
        sorting = se.NpzSortingExtractor(study_folder / 'ground_truth' / (rec_name + '.npz'))
        ground_truths[rec_name] = sorting
    return ground_truths


def run_study_sorters(study_folder, sorter_list, sorter_params={}, mode='keep',
                      engine='loop', engine_kwargs={}, verbose=False, run_sorter_kwargs={}):
    """
    Run all sorter on all recordings.


    Wrapper on top of st.sorter.run_sorters(...)


    Parameters
    ----------
    study_folder: str
        The study folder.

    sorter_params: dict of dict with sorter_name as key
        This allow to overwritte default params for sorter.

    mode: 'raise_if_exists' or 'overwrite' or 'keep'
        The mode when the subfolder of recording/sorter already exists.
            * 'raise' : raise error if subfolder exists
            * 'overwrite' : force recompute
            * 'keep' : do not compute again if f=subfolder exists and log is OK

    engine: str
        'loop', 'multiprocessing', or 'dask'

    engine_kwargs: dict
        This contains kwargs specific to the launcher engine:
            * 'loop' : no kargs
            * 'multiprocessing' : {'processes' : } number of processes
            * 'dask' : {'client':} the dask client for submiting task

    verbose: bool
        default True

    run_sorter_kwargs: dict
        This contains kwargs specific to run_sorter function:\
            * 'raise_error' :  bool
            * 'parallel' : bool
            * 'n_jobs' : int
            * 'joblib_backend' : 'loky' / 'multiprocessing' / 'threading'

    """
    study_folder = Path(study_folder)
    sorter_folders = study_folder / 'sorter_folders'

    recording_dict = get_recordings(study_folder)

    run_sorters(sorter_list, recording_dict, sorter_folders, sorter_params=sorter_params,
                grouping_property=None, mode=mode, engine=engine, engine_kwargs=engine_kwargs,
                with_output=False, verbose=verbose, run_sorter_kwargs=run_sorter_kwargs)

    # results are copied so the heavy sorter_folders can be removed
    copy_sortings_to_npz(study_folder)


def copy_sortings_to_npz(study_folder):
    """
    Collect sorting and copy then in npz format into a separate folder.
    Also copy

    """
    study_folder = Path(study_folder)
    sorter_folders = study_folder / 'sorter_folders'
    sorting_folders = study_folder / 'sortings'

    sorting_folders.mkdir(parents=True, exist_ok=True)

    for rec_name, sorter_name, output_folder in iter_output_folders(sorter_folders):
        SorterClass = sorter_dict[sorter_name]
        fname = rec_name + '[#]' + sorter_name
        npz_filename = sorting_folders / (fname + '.npz')
        try:
            sorting = SorterClass.get_result_from_folder(output_folder)
            se.NpzSortingExtractor.write_sorting(sorting, npz_filename)
        except:
            if npz_filename.is_file():
                npz_filename.unlink()
        if (output_folder / 'spikeinterface_log.json').is_file():
            shutil.copyfile(output_folder / 'spikeinterface_log.json', sorting_folders / 'run_log' / (fname + '.json'))


def iter_computed_names(study_folder):
    sorting_folder = Path(study_folder) / 'sortings'
    for filename in os.listdir(sorting_folder):
        if filename.endswith('.npz') and '[#]' in filename:
            rec_name, sorter_name = filename.replace('.npz', '').split('[#]')
            yield rec_name, sorter_name


def iter_computed_sorting(study_folder):
    """
    Iter over sorting files.
    """
    sorting_folder = Path(study_folder) / 'sortings'
    for filename in os.listdir(sorting_folder):
        if filename.endswith('.npz') and '[#]' in filename:
            rec_name, sorter_name = filename.replace('.npz', '').split('[#]')
            sorting = se.NpzSortingExtractor(sorting_folder / filename)
            yield rec_name, sorter_name, sorting


def collect_run_times(study_folder):
    """
    Collect run times in a working folder and sotre it in CVS files.

    The output is list of (rec_name, sorter_name, run_time)
    """
    study_folder = Path(study_folder)
    sorting_folders = study_folder / 'sortings'
    log_folder = sorting_folders / 'run_log'
    tables_folder = study_folder / 'tables'

    tables_folder.mkdir(parents=True, exist_ok=True)

    run_times = []
    for filename in os.listdir(log_folder):
        if filename.endswith('.json') and '[#]' in filename:
            rec_name, sorter_name = filename.replace('.json', '').split('[#]')
            with open(log_folder / filename, encoding='utf8', mode='r') as logfile:
                log = json.load(logfile)
                run_time = log.get('run_time', None)
            run_times.append((rec_name, sorter_name, run_time))

    run_times = pd.DataFrame(run_times, columns=['rec_name', 'sorter_name', 'run_time'])
    run_times = run_times.set_index(['rec_name', 'sorter_name'])

    return run_times


def aggregate_sorting_comparison(study_folder, exhaustive_gt=False):
    """
    Loop over output folder in a tree to collect sorting output and run
    ground_truth_comparison on them.

    Parameters
    ----------
    study_folder: str
        The study folder.
    exhaustive_gt: bool (default True)
        Tell if the ground true is "exhaustive" or not. In other world if the
        GT have all possible units. It allows more performance measurment.
        For instance, MEArec simulated dataset have exhaustive_gt=True

    Returns
    ----------
    comparisons: a dict of SortingComparison

    """

    study_folder = Path(study_folder)

    ground_truths = get_ground_truths(study_folder)
    results = collect_study_sorting(study_folder)

    comparisons = {}
    for (rec_name, sorter_name), sorting in results.items():
        gt_sorting = ground_truths[rec_name]
        sc = compare_sorter_to_ground_truth(gt_sorting, sorting, exhaustive_gt=exhaustive_gt)
        comparisons[(rec_name, sorter_name)] = sc

    return comparisons


def aggregate_performances_table(study_folder, exhaustive_gt=False, **karg_thresh):
    """
    Aggregate some results into dataframe to have a "study" overview on all recordingXsorter.

    Tables are:
      * run_times: run times per recordingXsorter
      * perf_pooled_with_sum: GroundTruthComparison.see get_performance
      * perf_pooled_with_average: GroundTruthComparison.see get_performance
      * count_units: given some threhold count how many units : 'well_detected', 'redundant', 'false_postive_units, 'bad'

    Parameters
    ----------
    study_folder: str
        The study folder.

    karg_thresh: dict
        Threholds paramerts used for the "count_units" table.

    Returns
    ----------

    dataframes: a dict of DataFrame
        Return several usefull DataFrame to compare all results.
        Note that count_units depend on karg_thresh.
    """
    study_folder = Path(study_folder)
    sorter_folders = study_folder / 'sorter_folders'
    tables_folder = study_folder / 'tables'

    comparisons = aggregate_sorting_comparison(study_folder, exhaustive_gt=exhaustive_gt)
    ground_truths = get_ground_truths(study_folder)
    results = collect_study_sorting(study_folder)

    study_folder = Path(study_folder)

    dataframes = {}

    # get run times:
    run_times = pd.read_csv(str(tables_folder / 'run_times.csv'), sep='\t')
    run_times.columns = ['rec_name', 'sorter_name', 'run_time']
    run_times = run_times.set_index(['rec_name', 'sorter_name', ])
    dataframes['run_times'] = run_times

    perf_pooled_with_sum = pd.DataFrame(index=run_times.index, columns=_perf_keys)
    dataframes['perf_pooled_with_sum'] = perf_pooled_with_sum

    perf_pooled_with_average = pd.DataFrame(index=run_times.index, columns=_perf_keys)
    dataframes['perf_pooled_with_average'] = perf_pooled_with_average

    count_units = pd.DataFrame(index=run_times.index,
                               columns=['num_gt', 'num_sorter', 'num_well_detected', 'num_redundant'])
    dataframes['count_units'] = count_units
    if exhaustive_gt:
        count_units['num_false_positive'] = None
        count_units['num_bad'] = None

    perf_by_spiketrain = []

    for (rec_name, sorter_name), comp in comparisons.items():
        gt_sorting = ground_truths[rec_name]
        sorting = results[(rec_name, sorter_name)]

        perf = comp.get_performance(method='pooled_with_sum', output='pandas')
        perf_pooled_with_sum.loc[(rec_name, sorter_name), :] = perf

        perf = comp.get_performance(method='pooled_with_average', output='pandas')
        perf_pooled_with_average.loc[(rec_name, sorter_name), :] = perf

        perf = comp.get_performance(method='by_spiketrain', output='pandas')
        perf['rec_name'] = rec_name
        perf['sorter_name'] = sorter_name
        perf = perf.reset_index()

        perf_by_spiketrain.append(perf)

        count_units.loc[(rec_name, sorter_name), 'num_gt'] = len(gt_sorting.get_unit_ids())
        count_units.loc[(rec_name, sorter_name), 'num_sorter'] = len(sorting.get_unit_ids())
        count_units.loc[(rec_name, sorter_name), 'num_well_detected'] = comp.count_well_detected_units(**karg_thresh)
        count_units.loc[(rec_name, sorter_name), 'num_redundant'] = comp.count_redundant_units()
        if exhaustive_gt:
            count_units.loc[(rec_name, sorter_name), 'num_false_positive'] = comp.count_false_positive_units()
            count_units.loc[(rec_name, sorter_name), 'num_bad'] = comp.count_bad_units()

    perf_by_spiketrain = pd.concat(perf_by_spiketrain)
    perf_by_spiketrain = perf_by_spiketrain.set_index(['rec_name', 'sorter_name', 'gt_unit_id'])
    dataframes['perf_by_spiketrain'] = perf_by_spiketrain

    return dataframes


def load_probe_file_inplace(recording, probe_file):
    '''
    This is a locally modified version of spikeextractor.extraction_tools.load_probe_file.
    But it do load "in place" and do NOT return a SubRecordingExtractor.
    
    This is usefull to not copy local raw data for KS, KS2, TDC, KLUSTA.
    
    This is a simplified version where there is only one group and **all** channel 
    of the file are in the groups.
    
    Work only for PRB file.
    
    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor to channel information
    probe_file: str
        Path to probe file. Either .prb or .csv
    verbose: bool
        If True, output is verbose

    Returns
    ---------
    
    Nothing, inplace modification of RecordingExtractor
    
    '''
    probe_file = Path(probe_file)
    
    assert probe_file.suffix == '.prb'
    
    
    probe_dict = read_python(probe_file)
    
    assert 'channel_groups' in probe_dict.keys()
    
    assert len(probe_dict['channel_groups']) == 1, 'load_probe_file_inplace only for one group'
    
    cgroup_id = list(probe_dict['channel_groups'].keys())[0]
    cgroup = probe_dict['channel_groups'][cgroup_id]
    
    channel_ids = cgroup['channels']
    assert len(channel_ids) == len(recording.get_channel_ids())
    # TODO assert equal array sorted
    
    for chan_id in channel_ids:
        recording.set_channel_property(chan_id, 'group', int(cgroup_id))
        recording.set_channel_property(chan_id, 'location', cgroup['geometry'][chan_id])







from pathlib import Path
import os

import numpy as np
import pandas as pd

import spikeextractors as se

from .comparisontools import _perf_keys
from .groundtruthcomparison import compare_sorter_to_ground_truth

from .studytools import (setup_comparison_study, run_study_sorters, get_rec_names,
                         get_one_recording, copy_sortings_to_npz, iter_computed_names,
                         iter_computed_sorting, collect_run_times)

import spiketoolkit as st


class GroundTruthStudy:
    def __init__(self, study_folder=None):
        self.study_folder = Path(study_folder)
        self._is_scanned = False
        self.computed_names = None
        self.rec_names = None
        self.sorter_names = None

        self.scan_folder()

        self.comparisons = None
        self.exhaustive_gt = None

    def __repr__(self):
        t = 'Groud truth study\n'
        t += '  ' + str(self.study_folder) + '\n'
        t += '  recordings: {} {}\n'.format(len(self.rec_names), self.rec_names)
        if len(self.sorter_names):
            t += '  sorters: {} {}\n'.format(len(self.sorter_names), self.sorter_names)

        return t

    def scan_folder(self):
        self.rec_names = get_rec_names(self.study_folder)
        # scan computed names
        self.computed_names = list(iter_computed_names(self.study_folder))  # list of pair (rec_name, sorter_name)
        self.sorter_names = np.unique([e for _, e in iter_computed_names(self.study_folder)]).tolist()
        self._is_scanned = True

    @classmethod
    def create(cls, study_folder, gt_dict):
        setup_comparison_study(study_folder, gt_dict)
        return cls(study_folder)

    def run_sorters(self, sorter_list, sorter_params={}, mode='keep', engine='loop', engine_kwargs={}, verbose=False,
                    run_sorter_kwargs={'parallel': False}):
        run_study_sorters(self.study_folder, sorter_list, sorter_params=sorter_params,
                          mode=mode, engine=engine, engine_kwargs=engine_kwargs, verbose=verbose,
                          run_sorter_kwargs=run_sorter_kwargs)

    def _check_rec_name(self, rec_name):
        if not self._is_scanned:
            self.scan_folder()
        if len(self.rec_names) > 1 and rec_name is None:
            raise Exception("Pass 'rec_name' parameter to select which recording to use.")
        elif len(self.rec_names) == 1:
            rec_name = self.rec_names[0]
        else:
            rec_name = self.rec_names[self.rec_names.index(rec_name)]
        return rec_name

    def get_ground_truth(self, rec_name=None):
        rec_name = self._check_rec_name(rec_name)
        sorting = se.NpzSortingExtractor(self.study_folder / 'ground_truth' / (rec_name + '.npz'))
        return sorting

    def get_recording(self, rec_name=None):
        rec_name = self._check_rec_name(rec_name)
        rec = get_one_recording(self.study_folder, rec_name)
        return rec

    def get_sorting(self, sort_name, rec_name=None):
        rec_name = self._check_rec_name(rec_name)

        selected_sorting = None
        if sort_name in self.sorter_names:
            for r_name, sorter_name, sorting in iter_computed_sorting(self.study_folder):
                if sort_name == sorter_name and r_name == rec_name:
                    selected_sorting = sorting
        return selected_sorting

    def copy_sortings(self):
        copy_sortings_to_npz(self.study_folder)
        self.scan_folder()

    def run_comparisons(self, exhaustive_gt=False, **kwargs):
        self.comparisons = {}
        for rec_name, sorter_name, sorting in iter_computed_sorting(self.study_folder):
            gt_sorting = self.get_ground_truth(rec_name)
            sc = compare_sorter_to_ground_truth(gt_sorting, sorting, exhaustive_gt=exhaustive_gt, **kwargs)
            self.comparisons[(rec_name, sorter_name)] = sc
        self.exhaustive_gt = exhaustive_gt

    def aggregate_run_times(self):
        return collect_run_times(self.study_folder)

    def aggregate_performance_by_units(self):
        assert self.comparisons is not None, 'run_comparisons first'

        perf_by_units = []
        for rec_name, sorter_name, sorting in iter_computed_sorting(self.study_folder):
            comp = self.comparisons[(rec_name, sorter_name)]

            perf = comp.get_performance(method='by_unit', output='pandas')
            perf['rec_name'] = rec_name
            perf['sorter_name'] = sorter_name
            perf = perf.reset_index()
            perf_by_units.append(perf)

        perf_by_units = pd.concat(perf_by_units)
        perf_by_units = perf_by_units.set_index(['rec_name', 'sorter_name', 'gt_unit_id'])

        return perf_by_units

    def aggregate_count_units(self, well_detected_score=None, redundant_score=None, overmerged_score=None):
        assert self.comparisons is not None, 'run_comparisons first'

        index = pd.MultiIndex.from_tuples(self.computed_names, names=['rec_name', 'sorter_name'])

        count_units = pd.DataFrame(index=index, columns=['num_gt', 'num_sorter', 'num_well_detected', 'num_redundant',
                                                         'num_overmerged'])

        if self.exhaustive_gt:
            count_units['num_false_positive'] = None
            count_units['num_bad'] = None

        for rec_name, sorter_name, sorting in iter_computed_sorting(self.study_folder):
            gt_sorting = self.get_ground_truth(rec_name)
            comp = self.comparisons[(rec_name, sorter_name)]

            count_units.loc[(rec_name, sorter_name), 'num_gt'] = len(gt_sorting.get_unit_ids())
            count_units.loc[(rec_name, sorter_name), 'num_sorter'] = len(sorting.get_unit_ids())
            count_units.loc[(rec_name, sorter_name), 'num_well_detected'] = \
                comp.count_well_detected_units(well_detected_score)
            if self.exhaustive_gt:
                count_units.loc[(rec_name, sorter_name), 'num_overmerged'] = \
                    comp.count_overmerged_units(overmerged_score)
                count_units.loc[(rec_name, sorter_name), 'num_redundant'] = \
                    comp.count_redundant_units(redundant_score)
                count_units.loc[(rec_name, sorter_name), 'num_false_positive'] = \
                    comp.count_false_positive_units(redundant_score)
                count_units.loc[(rec_name, sorter_name), 'num_bad'] = comp.count_bad_units()

        return count_units

    def aggregate_dataframes(self, copy_into_folder=True, **karg_thresh):
        dataframes = {}
        dataframes['run_times'] = self.aggregate_run_times().reset_index()
        perfs = self.aggregate_performance_by_units()

        dataframes['perf_by_units'] = perfs.reset_index()
        dataframes['count_units'] = self.aggregate_count_units(**karg_thresh).reset_index()

        if copy_into_folder:
            tables_folder = self.study_folder / 'tables'
            tables_folder.mkdir(parents=True, exist_ok=True)

            for name, df in dataframes.items():
                df.to_csv(str(tables_folder / (name + '.csv')), sep='\t', index=False)

        return dataframes

    def _compute_snr(self, rec_name, **snr_kargs):
        # print('compute SNR', rec_name)
        rec = self.get_recording(rec_name)
        gt_sorting = self.get_ground_truth(rec_name)

        snr_list = st.validation.compute_snrs(gt_sorting, rec, unit_ids=None, save_as_property=False, **snr_kargs)

        snr = pd.DataFrame(index=gt_sorting.get_unit_ids(), columns=['snr'])
        snr.index.name = 'gt_unit_id'
        snr.loc[:, 'snr'] = snr_list

        return snr

    def get_units_snr(self, rec_name=None, **snr_kargs):
        """
        Load or compute units SNR for a given recording.
        """
        rec_name = self._check_rec_name(rec_name)

        metrics_folder = self.study_folder / 'metrics'
        metrics_folder.mkdir(parents=True, exist_ok=True)

        filename = metrics_folder / ('SNR ' + rec_name + '.txt')

        if filename.is_file():
            snr = pd.read_csv(filename, sep='\t', index_col=None)
            snr = snr.set_index('gt_unit_id')
        else:
            snr = self._compute_snr(rec_name, **snr_kargs)
            snr.reset_index().to_csv(filename, sep='\t', index=False)
        snr['rec_name'] = rec_name
        return snr
    
    def concat_all_snr(self):
        snr = []
        for rec_name in self.rec_names:
            df = self.get_units_snr(rec_name)
            df = df.reset_index()
            snr.append(df)
        snr = pd.concat(snr)
        snr = snr.set_index(['rec_name', 'gt_unit_id'])
        return snr



from setuptools import setup, find_packages

d = {}
exec(open("spikecomparison/version.py").read(), None, d)
version = d['version']
long_description = open("README.md").read()

pkg_name = "spikecomparison"

setup(
    name=pkg_name,
    version=version,
    author="Alessio Buccino, Cole Hurwitz, Samuel Garcia, Jeremy Magland, Matthias Hennig",
    author_email="alessiop.buccino@gmail.com",
    description="Python toolkit for analysis, visualization, and comparison of spike sorting output",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alejoe91/spikecomparison",
    packages=find_packages(),
    package_data={},
    install_requires=[
        'numpy', 
        'scipy',
        'pandas',
        'networkx',
        'joblib',
        'spikeextractors>=0.9.3',
        'spiketoolkit>=0.7.2',
        'spikesorters>=0.4.3',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



import numpy as np
import spikeextractors as se

from .basecomparison import BaseTwoSorterComparison
from .comparisontools import make_possible_match, make_best_match, make_hungarian_match


class SymmetricSortingComparison(BaseTwoSorterComparison):
    """
    Class for symmetric comparison of two sorters when no assumption is done.
    """

    def __init__(self, sorting1, sorting2, sorting1_name=None, sorting2_name=None,
                 delta_time=0.4, sampling_frequency=None, match_score=0.5, chance_score=0.1,
                 n_jobs=-1, verbose=False):
        BaseTwoSorterComparison.__init__(self, sorting1, sorting2, sorting1_name=sorting1_name,
                                         sorting2_name=sorting2_name,
                                         delta_time=delta_time, sampling_frequency=sampling_frequency,
                                         match_score=match_score, chance_score=chance_score,
                                         n_jobs=n_jobs, verbose=verbose)

    def _do_matching(self):
        if self._verbose:
            print("Matching...")

        self.possible_match_12, self.possible_match_21 = make_possible_match(self.agreement_scores, self.chance_score)
        self.best_match_12, self.best_match_21 = make_best_match(self.agreement_scores, self.chance_score)
        self.hungarian_match_12, self.hungarian_match_21 = make_hungarian_match(self.agreement_scores,
                                                                                self.match_score)

    def get_mapped_sorting1(self):
        """
        Returns a MappedSortingExtractor for sorting 1.

        The returned MappedSortingExtractor.get_unit_ids returns the unit_ids of sorting 1.

        The returned MappedSortingExtractor.get_mapped_unit_ids returns the mapped unit_ids
        of sorting 2 to the units of sorting 1 (if units are not mapped they are labeled as -1).

        The returned MappedSortingExtractor.get_unit_spikeTrains returns the the spike trains
        of sorting 2 mapped to the unit_ids of sorting 1.
        """
        return MappedSortingExtractor(self.sorting2, self.hungarian_match_12.to_dict())

    def get_mapped_sorting2(self):
        """
        Returns a MappedSortingExtractor for sorting 2.

        The returned MappedSortingExtractor.get_unit_ids returns the unit_ids of sorting 2.

        The returned MappedSortingExtractor.get_mapped_unit_ids returns the mapped unit_ids
        of sorting 1 to the units of sorting 2 (if units are not mapped they are labeled as -1).

        The returned MappedSortingExtractor.get_unit_spikeTrains returns the the spike trains
        of sorting 1 mapped to the unit_ids of sorting 2.
        """
        return MappedSortingExtractor(self.sorting1, self.hungarian_match_21.to_dict())

    def get_matching_event_count(self, unit1, unit2):
        if (unit1 is not None) and (unit2 is not None):
            return self.match_event_count.at[unit1, unit2]
        else:
            raise Exception('get_matching_event_count: unit1 and unit2 must not be None.')

    def get_best_unit_match1(self, unit1):
        return self.best_match_12[unit1]

    def get_best_unit_match2(self, unit2):
        return self.best_match_21[unit2]

    def get_matching_unit_list1(self, unit1):
        return self.possible_match_12[unit1]

    def get_matching_unit_list2(self, unit2):
        return self.possible_match_21[unit2]

    def get_agreement_fraction(self, unit1=None, unit2=None):
        if unit1 is None or unit1 == -1 or unit2 is None or unit2 == -1:
            return 0
        else:
            return self.agreement_scores.at[unit1, unit2]


class MappedSortingExtractor(se.SortingExtractor):
    def __init__(self, sorting, unit_map):
        se.SortingExtractor.__init__(self)
        self._sorting = sorting
        self._unit_map = unit_map
        self._unit_ids = list(self._unit_map.keys())

    def get_unit_ids(self, unit_ids=None):
        if unit_ids is None:
            return self._unit_ids
        else:
            return self._unit_ids[unit_ids]

    def get_mapped_unit_ids(self, unit_ids=None):
        if unit_ids is None:
            return list(self._unit_map.values())
        elif isinstance(unit_ids, (int, np.integer)):
            return self._unit_map[unit_ids]
        else:
            return list([self._unit_map[u] for u in self._unit_ids if u in unit_ids])

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        unit2 = self._unit_map[unit_id]
        if unit2 != -1:
            return self._sorting.get_unit_spike_train(unit2, start_frame=start_frame, end_frame=end_frame)
        else:
            print(unit_id, " is not matched!")
            return None


def compare_two_sorters(sorting1, sorting2, sorting1_name=None, sorting2_name=None, delta_time=0.4,
                        sampling_frequency=None, match_score=0.5, chance_score=0.1, n_jobs=-1, verbose=False):
    '''
    Compares two spike sorter outputs.

    - Spike trains are matched based on their agreement scores
    - Individual spikes are labelled as true positives (TP), false negatives (FN), false positives 1 (FP from spike
      train 1), false positives 2 (FP from spike train 2), misclassifications (CL)

    It also allows to get confusion matrix and agreement fraction, false positive fraction and
    false negative fraction.

    Parameters
    ----------
    sorting1: SortingExtractor
        The first sorting for the comparison
    sorting2: SortingExtractor
        The second sorting for the comparison
    sorting1_name: str
        The name of sorter 1
    sorting2_name: : str
        The name of sorter 2
    delta_time: float
        Number of ms to consider coincident spikes (default 0.4 ms)
    sampling_frequency: float
        Optional sampling frequency in Hz when not included in sorting        
    match_score: float
        Minimum agreement score to match units (default 0.5)
    chance_score: float
        Minimum agreement score to for a possible match (default 0.1)
    n_jobs: int
        Number of cores to use in parallel. Uses all available if -1
    verbose: bool
        If True, output is verbose
    Returns
    -------
    sorting_comparison: SortingComparison
        The SortingComparison object

    '''
    return SymmetricSortingComparison(sorting1=sorting1, sorting2=sorting2, sorting1_name=sorting1_name,
                                      sorting2_name=sorting2_name, delta_time=delta_time,
                                      sampling_frequency=sampling_frequency,
                                      match_score=match_score, chance_score=chance_score,
                                      n_jobs=n_jobs, verbose=verbose)
